`timescale 1ns / 10ps
`default_nettype none

module tb_system;

   wire [7:0] led;
   reg       clk;
   reg       rst;
   wire       tx;
   reg       rx;

   system sys(
	      // Outputs
	      .tx_o			(tx),
	      .led_o			(led),
	      // Inputs
	      .clk_i			(clk),
	      .rx_i			(rx),
	      .rst_i			(rst));

   reg 	     done;
   
   initial begin
      done = 0;
      while(!done) begin
	 clk = 0;
	 #50;
	 clk = 1;
	 #50;
      end
   end


   // Well, we don't want to wait all day for the simulation
   // while sending commands over the UART in simulation...
   parameter BAUDRATE = 2;// Faster than this will not work!
   defparam  sys.mcu.io.uart.baudrate_r = BAUDRATE -1;


   reg [7:0] tx_char;

   task readchar;
     output reg [7:0] read_char;
     begin
	 @(posedge clk);
	 
	 while(tx == 1) begin
	    @(posedge clk);
	 end
	    
	 repeat (BAUDRATE / 2) @(posedge clk);
	 
	 repeat (8) begin
	    repeat (BAUDRATE) @(posedge clk);
	    tx_char <= (tx_char >> 1) | (tx<< 7);
	 end
	 
	 repeat (BAUDRATE) @(posedge clk);

	read_char <= tx_char;
	@(posedge clk);
	
     end
   endtask // readchar
      
   task expect_char;
      input [7:0] expectedchar;
      reg [7:0]   tmp;
      begin
	 readchar(tmp);
	 if(expectedchar !== tmp) begin
	    $display("%m: Did not receive the expected character: expected '%c' (%x), got '%c' (%x)",expectedchar,expectedchar,tx_char,tx_char);
	    $stop;
	 end
      end
   endtask // expect_char
   
      
   task wishbone_write;
      input [63:0] addr;
      input [63:0] data;
      begin
	 putch("m");
	 expect_char("m");
	 expect_char(" ");
	 write_longword(addr);
	 expect_char(" ");
	 write_longword(data);
	 expect_newline();
	 expect_char(">");
      end
   endtask // wishbone_write

   task expect_newline;
      begin
	 expect_char(8'h0d);
	 expect_char("\n");
      end
   endtask // expect_newline
   

   task wishbone_read;
      input [63:0] addr;
      output reg [63:0] data;
      begin
	 putch("d");
	 expect_char("d");
	 expect_char(" ");
	 write_longword(addr);
	 
	 expect_newline();
	 
	 read_longword(data);
	 
	   
	 expect_newline();
         expect_char(">");
      end
   endtask // wishbone_read
   

   task putch;
      input  [7:0] char;
      begin
	 $display("%m: Trying to write out '%c'",char);
	 @(posedge clk);

	 rx <= 0;
	 repeat(BAUDRATE) @(posedge clk);

	 rx <= char[0];
	 repeat(BAUDRATE) @(posedge clk);

	 rx <= char[1];
	 repeat(BAUDRATE) @(posedge clk);

	 rx <= char[2];
	 repeat(BAUDRATE) @(posedge clk);

	 rx <= char[3];
	 repeat(BAUDRATE) @(posedge clk);

	 rx <= char[4];
	 repeat(BAUDRATE) @(posedge clk);

	 rx <= char[5];
	 repeat(BAUDRATE) @(posedge clk);

	 rx <= char[6];
	 repeat(BAUDRATE) @(posedge clk);

	 rx <= char[7];
	 repeat(BAUDRATE) @(posedge clk);

	 // stopbit
	 rx <= 1;
	 repeat(BAUDRATE) @(posedge clk);
      end
   endtask // putch

   task bytetohex;
      input wire [7:0] c;
      output reg [3:0] x;
      begin
         if(c < 8'h30) begin
            $display("%m: Unexpected character %x",c);
            $stop;
         end if(c <= 8'h39) begin
            x = c - 8'h30;
         end else if(c < 8'h41) begin
            $display("%m: Unexpected character %x",c);
            $stop;
         end else if(c <= 8'h46) begin
            x = c - 8'h41 + 10;
         end else if(c < 8'h61) begin
            $display("%m: Unexpected character %x",c);
            $stop;
         end else if(c <= 8'h66) begin
            x = c - 8'h61 + 10;
         end else begin
            $display("%m: Unexpected character %x",c);
            $stop;
         end
      end
   endtask // bytetohex

   task write_longword;
      input reg [63:0] data;
      integer 	       i;
      begin
	 for(i = 7; i >= 0; i = i - 1) begin
	    putch(data[63:56]);
	    expect_char(data[63:56]);
	    data = data << 8;
	 end
      end
   endtask // write_longword


   
   task read_longword;
      output reg [31:0] x;
      reg [7:0]         c;
      reg [3:0]         i;
      begin
         for(i = 0; i < 8; i = i + 1) begin
            readchar(c);
            x = x << 4;
            bytetohex(c,x[3:0]);
         end
      end
   endtask // read_longword
   
   

   task confbus_write;
      input [63:0] addr;
      input [63:0] data;
      begin
	 wishbone_write(addr,data);
      end
   endtask // confbus_write

   reg [7:0] thecharacter;
   reg [31:0] doneval;
   reg [63:0] transactioncount;
   
   initial begin
      rx = 1;
      rst = 1;
      repeat(100) @(posedge clk);
      rst <= 0;

      thecharacter = 0;

      // Waiting for the prompt '>'      
      while(thecharacter != 8'h3e) begin 
	 readchar(thecharacter);
	 $display("%m: Got %c",thecharacter);
      end

      transactioncount = "00000500";
      
      confbus_write("00000000","00000000");
      confbus_write("00000004","3f000000");
      confbus_write("00000008","00000002");
      confbus_write("0000000c","0000003f");
      confbus_write("00000010","00000001");
      confbus_write("00000014",transactioncount);

      confbus_write("00000180","00000100");
      confbus_write("00000184","3f000000");
      confbus_write("00000188","00000002");
      confbus_write("0000018c","0000003f");
      confbus_write("00000190","00000001");
      confbus_write("00000194",transactioncount);

      confbus_write("00000300","00000200");
      confbus_write("00000304","3f000000");
      confbus_write("00000308","00000002");
      confbus_write("0000030c","0000003f");
      confbus_write("00000310","00000001");
      confbus_write("00000314",transactioncount);

      confbus_write("00000480","00000300");
      confbus_write("00000484","3f000000");
      confbus_write("00000488","00000002");
      confbus_write("0000048c","0000003f");
      confbus_write("00000490","00000001");
      confbus_write("00000494",transactioncount);

      confbus_write("00000600","00000400");
      confbus_write("00000604","3f000000");
      confbus_write("00000608","00000002");
      confbus_write("0000060c","0000003f");
      confbus_write("00000610","00000001");
      confbus_write("00000614",transactioncount);

      confbus_write("00000780","00000500");
      confbus_write("00000784","3f000000");
      confbus_write("00000788","00000002");
      confbus_write("0000078c","0000003f");
      confbus_write("00000790","00000001");
      confbus_write("00000794",transactioncount);

      confbus_write("00000900","00000600");
      confbus_write("00000904","3f000000");
      confbus_write("00000908","00000002");
      confbus_write("0000090c","0000003f");
      confbus_write("00000910","00000001");
      confbus_write("00000914",transactioncount);

      confbus_write("00000a80","00000700");
      confbus_write("00000a84","3f000000");
      confbus_write("00000a88","00000002");
      confbus_write("00000a8c","0000003f");
      confbus_write("00000a90","00000001");
      confbus_write("00000a94",transactioncount);

      confbus_write("00000c00","00000800");
      confbus_write("00000c04","3f000000");
      confbus_write("00000c08","00000002");
      confbus_write("00000c0c","0000003f");
      confbus_write("00000c10","00000001");
      confbus_write("00000c14",transactioncount);

      confbus_write("00000d80","00000900");
      confbus_write("00000d84","3f000000");
      confbus_write("00000d88","00000002");
      confbus_write("00000d8c","0000003f");
      confbus_write("00000d90","00000001");
      confbus_write("00000d94",transactioncount);

      confbus_write("00000f00","00000a00");
      confbus_write("00000f04","3f000000");
      confbus_write("00000f08","00000002");
      confbus_write("00000f0c","0000003f");
      confbus_write("00000f10","00000001");
      confbus_write("00000f14",transactioncount);

      confbus_write("00001080","00000b00");
      confbus_write("00001084","3f000000");
      confbus_write("00001088","00000002");
      confbus_write("0000108c","0000003f");
      confbus_write("00001090","00000001");
      confbus_write("00001094",transactioncount);
      


      // Start all masters:
      $display("All masters configured, sending start command");
      confbus_write("ffffffff","00000000");
      $display("Transactions started, waiting for everything to finish");

      while(!done) begin
	 wishbone_read("ffffffff",doneval);
	 
	 if(doneval == 32'h00000000) begin
	    $display("%m: All transaction generators finished");
	    done = 1;
	 end else if(doneval != 32'h00000001) begin
	    $display("Something went wrong, doneval is %x",doneval);
	    $stop;
	 end
      end
      
      
   end

endmodule // tb_system
