`timescale 1ns / 10ps
module transactiongenerator(
    input wire clk_i,rst_i,

    input wire [31:0] 	 confbus_i,
    output reg [31:0] 	 confbus_o,
    input wire 		 confbus_we_i,
    output reg 		 confbus_we_o,
    input wire [11:0]	 confbus_addr_i,
    output reg [11:0]	 confbus_addr_o,

    input wire 		 nextdata,
    input wire 		 nexttransaction,
    input wire 		 error, 		 
    output reg [31:0] 	 address,
    output reg [15:0]    length,
    output wire [31:0] 	 data_o,
    output reg 		 we,
    output reg 		 rd);
   


   parameter [6:0] 	 CONFADDR = 6'b0;

   localparam 		 C_ADDROR = 5'h0;
   localparam 		 C_ADDRAND = 5'h1;
   localparam 		 C_ADDRLENOR = 5'h2;
   localparam 		 C_ADDRLENAND = 5'h3;
   localparam 		 C_LFSR = 5'h4;
   localparam 		 C_TRANSNUM = 5'h5;
   localparam 		 C_STATUS = 5'h1f;
   
   


   reg [31:0] 		 addror_r;
   reg [31:0] 		 addrand_r;

   reg [15:0] 		 addrlenor_r;
   reg [15:0] 		 addrlenand_r;

   reg [15:0] 		 num_transactions_r;

   
   

   wire 		 matching_addr;
   assign 		 matching_addr = (CONFADDR == confbus_addr_i[11:5]) || // This address
					 (confbus_addr_i[11:5] == 7'h7f); // Also check for broadcast...

   reg 			 matching_addr_r;

   always @(posedge clk_i) begin
      matching_addr_r <= matching_addr;
   end


   wire [15:0] 		 status;
   
   always @(posedge clk_i) begin
      confbus_we_o <= confbus_we_i;
      confbus_addr_o <= confbus_addr_i;
      
      if(confbus_we_i) begin
	 confbus_o <= confbus_i;
      end else begin
	 confbus_o <= confbus_i;
	 if(matching_addr) begin
	    case(confbus_addr_i[4:0])
	      C_STATUS: confbus_o     <= {confbus_i[31:2],status[1:0]}  | confbus_i;
	    endcase // case(confbus_addr_i[5:0])
	 end // else: !if(matching_addr)
      end
   end // always @ (posedge clk_i)


   always @(posedge clk_i)
      if(rst_i) begin
	 addror_r <= 32'h0;
	 addrand_r <= 32'h0;
	 addrlenor_r <= 16'h0;
	 addrlenand_r <= 16'h0;
      end else begin
	 if( matching_addr_r && confbus_we_o) begin
	    case(confbus_addr_o[4:0])
	      C_ADDROR: begin addror_r <= confbus_o; $display("%m: Setting C_ADDROR to %x",confbus_o); end
	      C_ADDRAND: begin addrand_r <= confbus_o; $display("%m: Setting C_ADDRAND to %x",confbus_o); end
	      C_ADDRLENOR: begin addrlenor_r <= confbus_o; $display("%m: Setting C_ADDRLENOR to %x",confbus_o); end
	      C_ADDRLENAND: begin addrlenand_r <= confbus_o; $display("%m: Setting C_ADDRLENAND to %x",confbus_o); end
	      C_TRANSNUM: begin num_transactions_r <= confbus_o; $display("%m: Setting C_TRANSNUM to %x",confbus_o); end
	    endcase // case(confbus_addr_i[4:0])
	 end
      end // else: !if(rst_i)


   localparam LFSR_NEXT = 3'h0;
   localparam LFSR_LOAD = 3'h1;
   localparam LFSR_LOADSAVED = 3'h2;
   reg [2:0]  lfsrmux_r;
   reg [31:0] prevlfsr_r;
   reg [31:0] saved_lfsr_r; // The previous value of the lfsr register
   wire       loadsaved;
   reg 	      lfsrenable_r;
   wire [31:0] fromlfsr;


   always @(posedge clk_i) 
     if(rst_i) begin
	lfsrmux_r <= LFSR_NEXT;
     end else begin

	if(lfsrenable_r) begin
	   prevlfsr_r <= fromlfsr;
        end else if(confbus_we_o && matching_addr_r && (confbus_addr_o[4:0] == C_LFSR) )begin 
	   prevlfsr_r <= confbus_o;
	end

	// Note, do not care about nextdata here
	if(loadsaved) begin
	   lfsrmux_r <= LFSR_LOADSAVED;
	end else if(lfsrenable_r) begin
	   lfsrmux_r <= LFSR_NEXT; // Reset lfsrmux to default state
	end else if(confbus_we_o && matching_addr_r && (confbus_addr_o[4:0] == C_LFSR)) begin
	   lfsrmux_r <= LFSR_LOAD;
	end
     end // else: !if(rst_i)
   
   reg [31:0] tolfsr;

   always @*
     case(lfsrmux_r)
       LFSR_NEXT: tolfsr = fromlfsr;
       LFSR_LOAD: tolfsr = prevlfsr_r;
       LFSR_LOADSAVED: tolfsr = saved_lfsr_r;
       default: tolfsr = fromlfsr;
     endcase // case(lfsrmux_r)

   assign     data_o = fromlfsr;


   lfsr lfsr0(
	      // Outputs
	      .data_o			(fromlfsr),
	      // Inputs
	      .clk			(clk_i),
	      .enable			(lfsrenable_r || nextdata),
	      .data_i			(tolfsr));


   localparam  IDLE = 4'h0;
   localparam  TURNAROUND = 4'h1;
   localparam  GETADDR = 4'h2;
   localparam  GETLEN = 4'h3;
   localparam  GETDATA = 4'h4;
   localparam  GETREADADDR = 4'h5;
   localparam  GETREADLEN = 4'h6;
   localparam  GETREADDATA = 4'h7;
   localparam  READTURNAROUND = 4'h8;

   reg [3:0]   memtest_cs;

   reg [15:0]  curr_transaction;

   assign      loadsaved = (memtest_cs == GETDATA) && nexttransaction;

   always @(posedge clk_i)
     if(rst_i) begin
	memtest_cs <= IDLE;
	lfsrenable_r <= 1'b0;
	saved_lfsr_r <= 32'h0;
	curr_transaction <= 0;
     end else begin
	lfsrenable_r <= 0;
	we <= 1'b0;
	rd <= 1'b0;
	case(memtest_cs)
	  IDLE: begin
     if((matching_addr_r && confbus_we_o && (confbus_addr_o[4:0] == C_STATUS)) ) begin
		memtest_cs <= TURNAROUND;
		lfsrenable_r <= 1'b1;
		curr_transaction <= num_transactions_r;
	     end
	  end

	  TURNAROUND: begin
	     memtest_cs <=  GETADDR;
	     lfsrenable_r <= 1'b1;
	     saved_lfsr_r <= tolfsr;
	  end


	  GETADDR: begin
	     memtest_cs <= GETLEN;
	     address <= (fromlfsr & addrand_r) | addror_r;
	     lfsrenable_r <= 1'b1;
	  end

	  GETLEN: begin
	     memtest_cs <= GETDATA;
	     length <= (fromlfsr & addrlenand_r) | addrlenor_r;
	     lfsrenable_r <= 1'b1;
	     we <= 1'b1;
	  end

	  GETDATA: begin
	     if(nexttransaction) begin
		// In this state loadsaved will also be activated
		// combinatorially above
		memtest_cs <= READTURNAROUND;
		lfsrenable_r <= 1'b1;
	     end
	  end

	  READTURNAROUND: begin
	     lfsrenable_r <= 1'b1;
	     memtest_cs <= GETREADADDR;
	  end

	  GETREADADDR: begin
	     memtest_cs <= GETREADLEN;
	     address <= (fromlfsr & addrand_r) | addror_r;
	     lfsrenable_r <= 1'b1;
	  end

	  GETREADLEN: begin
	     memtest_cs <= GETREADDATA;
	     length <= (fromlfsr & addrlenand_r) | addrlenor_r;
	     lfsrenable_r <= 1'b1;
	     rd <= 1'b1;
	  end

	  GETREADDATA: begin
	     if(nexttransaction && (curr_transaction != 1) ) begin
		// In this state loadsaved will also be activated
		// combinatorially above
		curr_transaction <= curr_transaction - 1;
		memtest_cs <= TURNAROUND;
		lfsrenable_r <= 1'b1;
		$display("%m: New transaction with seed %08x\n",fromlfsr);

	     end else if(nexttransaction) begin
		$display("%m: Finished all transactions\n");
		memtest_cs <= IDLE;
	     end
	  end

	endcase // case(memtest_cs)
     end // else: !if(rst_i)

   assign status = {14'h0, error, (memtest_cs == IDLE) ? 1'b0 : 1'b1};
   
endmodule // transactiongenerator
