`default_nettype none
`timescale 1ns / 10ps

// In order to keep the performance high we need flip flops
// on both the inputs and outputs of the block ram
// This means that we get a delay of 3 clock cycles before we
// get any data when trying to read from the block ram...
// Writes can be accepted immediately however
module wb_slave(input wire clk_i,rst_i,

    // Wishbone slave port
    input wire           wbs_stb_i,
    input wire 		 wbs_cyc_i,
    input wire 		 wbs_cab_i,
    input wire 		 wbs_we_i,
    input wire [31:0] 	 wbs_adr_i,
    input wire [31:0] 	 wbs_dat_i,
    input wire [3:0] 	 wbs_sel_i,
    output reg [31:0] 	 wbs_dat_o,
    output reg  	 wbs_ack_o,
    output wire  	 wbs_rty_o,
    output wire  	 wbs_err_o
		);




   localparam [1:0] 	 IDLE = 2'd0;
   localparam [1:0] 	 GOTSTB = 2'd1;
   localparam [1:0] 	 READY = 2'd2;
   
   reg [1:0]		 state;

   assign 		 wbs_rty_o = 0;
   assign 		 wbs_err_o = 0;
   wire [11:0] next_readaddr;
   

   
   reg [11:0] 		 readaddr;


   reg 			 stb, cyc, cab, we;
   reg [11:0] 		 adr;
   reg [31:0] 		 dat_i;
   reg [3:0] 		 sel_i;

   always @(posedge clk_i) begin
      if(rst_i) begin
	 stb   <= 0;
	 cyc   <= 0;
	 adr   <= 0;
	 cab   <= 0;
	 we    <= 0;
	 dat_i <= 0;
	 sel_i <= 0;
      end else begin
	 stb <= wbs_stb_i;
	 cyc <= wbs_cyc_i;
	 cab <= wbs_cab_i;
	 we <= wbs_we_i;
	 adr <= wbs_adr_i[13:2];
	 dat_i <= wbs_dat_i;
	 sel_i <= wbs_sel_i;
      end
   end

   /*
    clk    ¯¯|__|¯¯|__|¯¯|__|¯¯|__|¯¯|__|¯¯|__|
    stb_i  ___|¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯|________
    stb    _________|¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯|__
    ack_o  __________________|¯¯¯¯¯¯¯¯¯¯¯¯¯¯|__
    */

   always @(posedge clk_i) begin
      // This memory assumes that it has 18 address bits allocated to it
      // and that 16 of these will wrap around as normal.
      // However, it will complain if 17:16 is nonzero at a transaction.
      // (This has trapped a couple of bugs where the address part of a NoC
      //  message was not read correctly.)
      if(wbs_adr_i[17:16] && wbs_stb_i && wbs_cyc_i) begin
	 $display("%m: Illegal addr 0x%x",wbs_adr_i);
	 $stop;
      end

      if(wbs_stb_i && wbs_cyc_i) begin
	 
	 if(wbs_we_i) begin
	    if(wbs_sel_i != 4'hf) begin
	       $display("%m: Error, does not handle wbs_sel_i != 1111 at address 0x%08x",wbs_adr_i);
	       $stop;
	    end

	    // If you need to debug something this might be nice to uncomment
//	    if(wbs_ack_o) begin
//	       $display("%m: Wrote %x to memory location %x (/4)",wbs_dat_i,wbs_adr_i);
//	    end

	 end
      end
   end // always @ (posedge clk_i)

   wire [11:0] memaddr;
   assign      memaddr = state ? readaddr : adr;

   wire        en = stb && cyc;

   wire [31:0]  foo;
   always @(posedge clk_i) begin
      wbs_dat_o <= foo;
   end
   
   RAMB16_S4 ram7
     (
      .DO(foo[31:28]),
      .ADDR(memaddr[11:0]),
      .CLK(clk_i),
      .DI(dat_i[31:28]),
      .EN(en &&  sel_i[3]),
      .SSR(1'b0),
      .WE(we)
      );

   RAMB16_S4 ram6
     (
      .DO(foo[27:24]),
      .ADDR(memaddr[11:0]),
      .CLK(clk_i),
      .DI(dat_i[27:24]),
      .EN(en &&  sel_i[3]),
      .SSR(1'b0),
      .WE(we)
      );

   RAMB16_S4 ram5
     (
      .DO(foo[23:20]),
      .ADDR(memaddr[11:0]),
      .CLK(clk_i),
      .DI(dat_i[23:20]),
      .EN(en &&  sel_i[2]),
      .SSR(1'b0),
      .WE(we)
      );
   RAMB16_S4 ram4
     (
      .DO(foo[19:16]),
      .ADDR(memaddr[11:0]),
      .CLK(clk_i),
      .DI(dat_i[19:16]),
      .EN(en &&  sel_i[2]),
      .SSR(1'b0),
      .WE(we)
      );

   RAMB16_S4 ram3
     (
      .DO(foo[15:12]),
      .ADDR(memaddr[11:0]),
      .CLK(clk_i),
      .DI(dat_i[15:12]),
      .EN(en &&  sel_i[1]),
      .SSR(1'b0),
      .WE(we)
      );

   RAMB16_S4 ram2
     (
      .DO(foo[11:8]),
      .ADDR(memaddr[11:0]),
      .CLK(clk_i),
      .DI(dat_i[11:8]),
      .EN(en &&  sel_i[1]),
      .SSR(1'b0),
      .WE(we)
      );

   RAMB16_S4 ram1
     (
      .DO(foo[7:4]),
      .ADDR(memaddr[11:0]),
      .CLK(clk_i),
      .DI(dat_i[7:4]),
      .EN(en &&  sel_i[0]),
      .SSR(1'b0),
      .WE(we)
      );
   RAMB16_S4 ram0
     (
      .DO(foo[3:0]),
      .ADDR(memaddr[11:0]),
      .CLK(clk_i),
      .DI(dat_i[3:0]),
      .EN(en &&  sel_i[0]),
      .SSR(1'b0),
      .WE(we)
      );


   wire        selector;

   assign      selector = (state != IDLE);

   wire [11:0] tmp;

   assign      tmp = selector ? readaddr : adr;
   assign      next_readaddr = tmp + 1;


   always @(posedge clk_i) begin
      if(rst_i) begin
	 state <= IDLE;
      end else begin

	 readaddr <= next_readaddr;
	 wbs_ack_o <= 0;

	 case(state)
	   IDLE: begin
	    if(!we && stb && cyc) begin
	       state <= GOTSTB;
	    end

	      // Writes can be acked immediately
	    if(we && stb && cyc && wbs_stb_i && wbs_cyc_i) begin
	       wbs_ack_o <= 1;
	    end
	      
	   end // case: IDLE

	   GOTSTB: begin
	      state <= READY;
	      wbs_ack_o <= 1; // Read will be ready next cycle...
	   end

	   READY: begin
	      if(stb && cyc && wbs_stb_i && wbs_cyc_i) begin
		 wbs_ack_o <= 1;
	      end

	      if(!we && stb && cyc && cab) begin
		 state <= READY;
	      end else begin
		 state <= IDLE;
	      end
	   end
	     
	 endcase // case(state)
      end // else: !if(rst_i)
   end // always @ (posedge clk_i)
   
   
endmodule // wb_slave
